#include<stdio.h>
#include<cstdlib>
#include<complex>

using namespace std;

#include"uppe_solver.h"

const int NX = 500;

// these PML-related arrays will hold the coefficient that modify the 1D Laplacian
complex<double> AA[NX];
complex<double> BB[NX];

// structure that trasnports parameters to ODE solver [= void *p in RHS_function signature]
class RHSpars {
public:
  int    nc;
  double k0;
  double dr;
};

int RHS_function(double z, const double y[], double dy[], void *p) {
  RHSpars *pars = (RHSpars*) p;
  int      nc = pars->nc;
  double   k0 = pars->k0;
  double   dr = pars->dr;

  complex<double>  idelta = complex<double>(0.0, 1.0/(2*k0*dr*dr) );
  complex<double> *am    = (complex<double>*)  y;
  complex<double> *da    = (complex<double>*) dy;
  double          dxcoef = 1.0/2.0/dr;

  
  // PML-related modification in AA,BB
  for(int r=1;r<nc-1;r++) {
    da[r] = idelta*(am[r-1] -2.0*am[r] + am[r+1])*AA[r] + BB[r]*(am[r+1] - am[r-1])*dxcoef;
  }
  
  // the "real" boundary conditon - PEC or periodic
  da[0]    = 1.0*idelta*(am[nc-1] - 2.0*am[0]    + am[1]);
  da[nc-1] = 1.0*idelta*(am[nc-2] - 2.0*am[nc-1] + am[0]);

  return(0);
}


complex<double> gaussx(double wr, double r, double z, double k, double f, double kx) 
{
  complex<double> aux1, aux2, icko(0.0,1.0);

  if( wr <= 0.0 ) return(1.0);

  if(f != 0.0) 
    aux1 =  1.0/(wr*wr) + icko*k/(2.0*f);
  else
    aux1 =  1.0/(wr*wr);
  aux2 =  1.0/(1.0 + 2.0*icko*z/k*aux1);

  return( sqrt(aux2)*exp(-r*r*aux1*aux2 + complex<double>(0.0,kx*r)));
}





int main(int argc, char *argv[]) {
  int x;

  Solver  solver;
  RHSpars rpars;

  /* variables/values as in the matlab scritp */
  double    lambda = 633.0e-09;          
  double    w0     = 1.0e-04;
  double    angle = +1.0;

  /* computational domain */
  double    LX     = 0.25e-02;
  double    z      = 0.0;
  double    dz     = 0.001;
  double    LZ     = 0.3;

  /* derived parameters */
  double k0   = 2*M_PI/lambda;
  double dx   = LX/NX;
  double kx   = k0*angle/360.0*2*M_PI;

  /* coordinates */
  double cx[NX];
  for(x=0;x<NX;x++) cx[x] = dx*(x-NX/2);

  /* initial amplitude */
  complex<double> a0[NX];
  for(x=0;x<NX;x++) a0[x] = gaussx(w0, cx[x], 0.0, k0, 0.0, +kx);

  /* working amplitude */
  complex<double> am[NX];
  for(x=0;x<NX;x++) am[x] = a0[x];

  /* to send parameters to RHSfunction */
  rpars.nc = NX;
  rpars.dr = dx;
  rpars.k0 = k0;

  /* intialize ODE solver */
  solver.Init("rk4",
             1.0e-10,
             1.0e-09,
             &rpars,
             NX*2,
             RHS_function,
             0,
             true );


  /************************** initailize PML ******************************************************/
  int n0;
  complex<double> I = complex<double>(0.0,1.0);
  double ww = 1.0/( LX/4.0 );

  /* these values = no PML */
  for(x=0;x<NX;x++) AA[x] = 1.0;
  for(x=0;x<NX;x++) BB[x] = 0.0;

  /* implementing formulas from lecture notes        */
  /* n0 = index separating inner domain from the PML */

  /* 15 percent to PML */
  n0 = floor(0.85*NX);
  for(x=n0;x<NX;x++) {
    AA[x] = -1.0/pow(I - 4.0*pow(ww*(cx[x] - cx[n0]),3.0),2.0);
    BB[x] = -6.0*I*ww*ww*ww*pow(cx[x]-cx[n0],2.0)/(k0*pow(I - 4.0*pow(ww*(cx[x] - cx[n0]),3.0),3.0));
  }

  /* 15 percent to PML */
  n0 = floor(0.15*NX);
  for(x=0;x<n0;x++) {
    AA[x] = -1.0/pow(I + 4.0*pow(ww*(cx[x] - cx[n0]),3.0),2.0);
    BB[x] = +6.0*I*ww*ww*ww*pow(cx[x]-cx[n0],2.0)/(k0*pow(I + 4.0*pow(ww*(cx[x] - cx[n0]),3.0),3.0));
  }

  /*************************  end PML adition *****************************************************/

  /* file to save results */
  FILE *fout;
  fout = fopen("LogIMap.dat","w");


  /* evolve amplitude */
  int k,n;  
  for(n=0;n<500;n++) {
    // split main loop into 200 chunks 
    z = 0.0;
    for(k=0;z<LZ/500.0;k++) {
      solver.Step_adaptive(&z,LZ/500.0,&dz,(double*) am);
    }

    // after each chunk finishes, save beam profile
    for(x=0;x<NX;x++) fprintf(fout,"%g ", log10(norm(am[x]) + 1.0e-20 ));
    fprintf(fout,"\n");

    cout << n <<" "<< dz  << endl;
  }
  
  fclose(fout);

  return(0);
}

