#include<stdio.h>
#include<cstdlib>
#include<complex>

using namespace std;

#include"uppe_solver.h"

const int NX = 250;

// these PML-related arrays will hold the coefficient that modify the 1D Laplacian
complex<double> AA[NX];
complex<double> BB[NX];

// structure that trasnports parameters to ODE solver [= void *p in RHS_function signature]
class RHSpars {
public:
  int    nc;
  double k0;
  double dr;
};

int RHS_function(double z, const double yy[], double dy[], void *p) {
  RHSpars *pars = (RHSpars*) p;
  int      nc = pars->nc;
  double   k0 = pars->k0;
  double   dr = pars->dr;

  complex<double>  idelta = complex<double>(0.0, 1.0/(2*k0*dr*dr) );
  complex<double> *am    = (complex<double>*) yy;
  complex<double> *da    = (complex<double>*) dy;
  double          dxcoef = 1.0/2.0/dr;

  int x,y;
  for(x=0;x<nc*nc;x++) da[x] = 0.0;
  
  // PML-related modification in AA,BB
  for(y=1;y<nc-1;y++) {
    for(x=1;x<nc-1;x++) {
      da[x + y*nc]  = idelta*(am[x-1 + y*nc    ] -2.0*am[x + y*nc] + am[x+1 + y*nc    ])*AA[x] + BB[x]*(am[x+1 + y*nc    ] - am[x-1 + y*nc    ])*dxcoef;
      da[x + y*nc] += idelta*(am[x   + (y-1)*nc] -2.0*am[x + y*nc] + am[x   + (y+1)*nc])*AA[y] + BB[y]*(am[x   + (y+1)*nc] - am[x   + (y-1)*nc])*dxcoef;
    }
  }

  // the "real" boundary conditon - PEC or periodic
    //  da[0]    = 1.0*idelta*(am[nc-1] - 2.0*am[0]    + am[1]);
    //  da[nc-1] = 1.0*idelta*(am[nc-2] - 2.0*am[nc-1] + am[0]);

  return(0);
}


complex<double> gaussx(double wr, double r, double z, double k, double f, double kx) 
{
  complex<double> aux1, aux2, icko(0.0,1.0);

  if( wr <= 0.0 ) return(1.0);

  if(f != 0.0) 
    aux1 =  1.0/(wr*wr) + icko*k/(2.0*f);
  else
    aux1 =  1.0/(wr*wr);
  aux2 =  1.0/(1.0 + 2.0*icko*z/k*aux1);

  return( sqrt(aux2)*exp(-r*r*aux1*aux2 + complex<double>(0.0,kx*r)));
}





int main(int argc, char *argv[]) {
  int x,y;

  Solver  solver;
  RHSpars rpars;

  /* variables/values as in the matlab scritp */
  double    lambda = 633.0e-09;          
  double    w0     = 1.0e-04;
  double    angle = +1.0;

  /* computational domain */
  double    LX     = 0.0625e-02;
  double    z      = 0.0;
  double    dz     = 1.0e-06;
  double    LZ     = 0.3;

  /* derived parameters */
  double k0   = 2*M_PI/lambda;
  double dx   = LX/NX;
  double kx   = k0*angle/360.0*2*M_PI;

  /* coordinates */
  double cx[NX];
  for(x=0;x<NX;x++) cx[x] = dx*(x-NX/2 + 0.5);

  /* initial amplitude */
  complex<double> ax[NX];
  for(x=0;x<NX;x++) ax[x] = gaussx(w0, cx[x], 0.0, k0, 0.0, +kx);

  complex<double> ay[NX];
  for(x=0;x<NX;x++) ay[x] = gaussx(w0, cx[x], 0.0, k0, 0.0, +kx);

  /* working amplitude */
  complex<double> *am = new complex<double>[NX*NX];
  for(x=0;x<NX;  x++) for(y=0;y<NX;  y++) am[x+y*NX] = 0.0;
  for(x=1;x<NX-1;x++) for(y=1;y<NX-1;y++) am[x+y*NX] = ax[x]*ay[y];

  /* to send parameters to RHSfunction */
  rpars.nc = NX;
  rpars.dr = dx;
  rpars.k0 = k0;

  /* intialize ODE solver */
  solver.Init("rkf45",
             1.0e-09,
             0.0e-09,
             &rpars,
             NX*NX*2,
             RHS_function,
             0,
             true );


  /************************** initailize PML ******************************************************/
  int n0;
  complex<double> I = complex<double>(0.0,1.0);
  double ww = 1.0/( LX/6.0 );

  /* these values = no PML */
  for(x=0;x<NX;x++) AA[x] = 1.0;
  for(x=0;x<NX;x++) BB[x] = 0.0;

  /* implementing formulas from lecture notes        */
  /* n0 = index separating inner domain from the PML */

  /* 15 percent to PML */
  n0 = floor(0.85*NX);
  for(x=n0;x<NX;x++) {
    AA[x] = -1.0/pow(I - 4.0*pow(ww*(cx[x] - cx[n0]),3.0),2.0);
    BB[x] = -6.0*I*ww*ww*ww*pow(cx[x]-cx[n0],2.0)/(k0*pow(I - 4.0*pow(ww*(cx[x] - cx[n0]),3.0),3.0));
  }

  /* 15 percent to PML */
  n0 = floor(0.15*NX);
  for(x=0;x<n0;x++) {
    AA[x] = -1.0/pow(I + 4.0*pow(ww*(cx[x] - cx[n0]),3.0),2.0);
    BB[x] = +6.0*I*ww*ww*ww*pow(cx[x]-cx[n0],2.0)/(k0*pow(I + 4.0*pow(ww*(cx[x] - cx[n0]),3.0),3.0));
  }


  /*************************  end PML adition *****************************************************/

  /* file to save results */
  FILE *fout;



  /* evolve amplitude */
  int k,n;  
  for(n=0;n<5;n++) {
    // split main loop into X chunks 
    z = 0.0;
    for(k=0;z<LZ/100.0;k++) {
      solver.Step_adaptive(&z,LZ/100.0,&dz,(double*) am);
    }

    // after each chunk finishes, save beam profile

    fout = fopen("Map.dat","w");
    for(y=0;y<NX;y++) {
      for(x=0;x<NX;x++) fprintf(fout,"%g ", real(am[x + NX*y]) );
      //      for(x=0;x<NX;x++) fprintf(fout,"%g ", log10(norm(am[x + NX*y]) + 1.0e-20 ));
      fprintf(fout,"\n");
    }
    fclose(fout);

    cout << n <<" "<< dz  << endl;
  }
  


  return(0);
}

